<?php

require_once 'phing/Task.php';
require_once 'Text/Diff.php';
require_once 'Text/Diff/Renderer/unified.php';
require_once 'HTTP/Request2.php';

/**
 * Phing MODx patch task
 *
 * @author Michiel Rook <info@touchdownconsulting.nl>
 * @version $Id: ModxChangedTask.php 288 2010-03-24 19:51:59Z michiel $
 */
class ModxChangedTask extends Task
{
    const ACTION_SAVE_DOCUMENT = 5;
    const ACTION_SAVE_TEMPLATE = 20;
    const ACTION_SAVE_SNIPPET = 24;
    const ACTION_SAVE_CHUNK = 79;
    
    private $hostname = "localhost";
    private $database = "database";
    private $username = "";
    private $password = "";
    private $prefix = "";
    private $timestamp = 0;
    private $remoteUrl = "";
    private $remoteKey = "";
    private $db = null;
    
    private $changes = array();
    private $sqlScript = "";
    
    /**
     * Sets hostname
     * @param string $hostname
     */
    public function setHostname($hostname)
    {
        $this->hostname = $hostname;
    }
    
    /**
     * Sets database
     * @param string $database
     */
    public function setDatabase($database)
    {
        $this->database = $database;
    }

    /**
     * Sets username
     * @param string $username
     */
    public function setUsername($username)
    {
        $this->username = $username;
    }

    /**
     * Sets password
     * @param string $password
     */
    public function setPassword($password)
    {
        $this->password = $password;
    }

    /**
     * Sets prefix
     * @param string $prefix
     */
    public function setPrefix($prefix)
    {
        $this->prefix = $prefix;
    }

    /**
     * Sets timestamp
     * @param string $timestamp
     */
    public function setTimestamp($timestamp)
    {
        $this->timestamp = strtotime($timestamp);
        
        if ($this->timestamp === false) {
            $this->timestamp = (int) $timestamp;
        }
    }

    /**
     * Sets remote URL
     * @param string $remoteUrl
     */
    public function setRemoteUrl($remoteUrl)
    {
        $this->remoteUrl = $remoteUrl;
    }

    /**
     * Sets remote key
     * @param string $remoteKey
     */
    public function setRemoteKey($remoteKey)
    {
        $this->remoteKey = $remoteKey;
    }

    public function main()
    {
        $this->db = new PDO(
            "mysql:host={$this->hostname};dbname={$this->database}",
            $this->username, 
            $this->password, 
            array(PDO::MYSQL_ATTR_INIT_COMMAND => "SET NAMES utf8") 
        );
        
        echo "Searching changes from " . date('Y-m-d', $this->timestamp) . ":\n\n";
        
        $this->findChanges();
        
        foreach ($this->changes as $key => $changes) {
            $this->compileChanges($key, $changes);
        }
        
        file_put_contents('modx_migrate.sql', $this->sqlScript);
    }
    
    private function findChanges()
    {
        $actions = array(
            self::ACTION_SAVE_CHUNK,
            self::ACTION_SAVE_DOCUMENT,
            self::ACTION_SAVE_SNIPPET,
            self::ACTION_SAVE_TEMPLATE
        );
        
        $actionsImploded = implode(',', $actions);
        
        $statement = $this->db->prepare("SELECT FROM_UNIXTIME(timestamp,'%Y-%m-%d %H:%i:%S') AS date,action,itemid,message
        FROM {$this->prefix}manager_log WHERE action IN ({$actionsImploded}) AND timestamp > :timestamp ORDER BY timestamp");
        
        $statement->bindParam(':timestamp', $this->timestamp, PDO::PARAM_INT);
        
        if ($statement->execute()) {
            while ($row = $statement->fetch()) {
                $action = (int) $row['action'];
                $itemid = (int) $row['itemid'];
                $date = $row['date'];
                
                if (!$itemid) {
                    continue;
                }
                
                switch ($action) {
                    case self::ACTION_SAVE_CHUNK: {
                        $this->changes['chunks'][$itemid] = $date;
                    } break;

                    case self::ACTION_SAVE_DOCUMENT: {
                        $this->changes['documents'][$itemid] = $date;
                    } break;

                    case self::ACTION_SAVE_SNIPPET: {
                        $this->changes['snippets'][$itemid] = $date;
                    } break;

                    case self::ACTION_SAVE_TEMPLATE: {
                        $this->changes['templates'][$itemid] = $date;
                    } break;
                }
            }
        } else {
            throw new BuildException("Could not get database info");
        }
    }

    private function getDump($type)
    {
        $request = new HTTP_Request2();
        $request->setAdapter('curl');
        
        $request->setUrl($this->remoteUrl . 'updserver/server.php');
        
        $url = $request->getUrl();
        $url->setQueryVariables(array(
            'token' => md5($this->remoteKey),
            'cmd'   => $type
        ));
        
        try {
            return $request->send()->getBody();
        } catch (HTTP_Request2_Exception $e) {
            throw new BuildException($e->getMessage());
        }
    }
    
    private function compileChanges($type, $changes)
    {
        $header = ucfirst($type);
        $column = '';
        $table = '';
        
        switch ($type) {
            case 'documents': {
                $column = 'pagetitle';
                $contentColumn = 'content';
                $table = $this->prefix . 'site_content';
            } break;
            
            case 'snippets': {
                $column = 'name';
                $contentColumn = 'snippet';
                $table = $this->prefix . 'site_snippets';
            } break;
            
            case 'templates': {
                $column = 'templatename';
                $contentColumn = 'content';
                $table = $this->prefix . 'site_templates';
            } break;
            
            case 'chunks': {
                $column = 'name';
                $contentColumn = 'snippet';
                $table = $this->prefix . 'site_htmlsnippets';
            } break;
        }
        
        echo "===== " . str_pad($header . ' ', 49, '=') . "\n";

        $contents = $this->getDump($type);
        
        $tmptable = 'tmp_' . $table;
        
        $contents = str_replace($table, $tmptable, $contents);
        
        $statement = $this->db->prepare($contents);
        if (!$statement->execute()) {
            $message = $statement->errorInfo();
            echo $message[2] . "\n";
        }
        
        $newItems = array();
        
        $statement = $this->db->prepare("SELECT * FROM {$table}");
        
        if ($statement->execute()) {
            while ($row = $statement->fetch()) {
                /*$doc = array(
                    'longtitle' => $row['longtitle'],
                    'pagetitle' => $row['pagetitle'],
                    'alias' => $row['alias'],
                    'content' => $row['content'],
                    'status' => 'new'
                );*/
                
                $newItems[$row['id']] = true;
            }
        }
        
        $statement = $this->db->prepare("SHOW COLUMNS FROM {$table}");
        $statement->execute();

        $fields = array();

        while ($row = $statement->fetch()) {
            $fields[] = $row[0];
        }
        
        $statement = $this->db->prepare("SELECT * FROM {$tmptable} WHERE id = :id");
        
        foreach ($newItems as $id => $doc) {
            $statement->bindValue(':id', $id);
            $statement->execute();
            
            $row = $statement->fetch();
            
            if ($row === false) {
                $changes[$id] = date('Y-m-d H:i:s');
            }
        }

        $changeCount = 0;
        $newStatement = $this->db->prepare("SELECT * FROM {$table} WHERE id = :id");
        
        foreach ($changes as $id => $date) {
            $newStatement->bindValue(':id', $id);
            $newStatement->execute();
            $newRow = $newStatement->fetch();

            $statement->bindValue(':id', $id);
            $statement->execute();
            $row = $statement->fetch();

            if ($row === false) {
                $status = 'NEW';
            } else {
                if (!strcmp($row[$contentColumn], $newRow[$contentColumn])) {
                    continue;
                }
                
                $status = 'UPD';
            }
            
            $linesOrig = explode("\n", $row[$contentColumn]);
            $linesNew = explode("\n", $newRow[$contentColumn]);
                
            $diff = new Text_Diff($linesOrig, $linesNew);
            $renderer = new Text_Diff_Renderer_unified();
            
            echo "[" . str_pad($id, 3, ' ', STR_PAD_LEFT) . "] " . str_pad($newRow[$column], 30) . "{$date} [{$status}]\n";
            
            if ($status == 'UPD')
                echo $renderer->render($diff) . "\n";
            
            $values = array();
            
            foreach ($fields as $field) {
                $val = addcslashes($newRow[$field], "\n\r'");
                
                if (!is_numeric($val)) {
                    $val = "'" . $val . "'";
                }
                
                $values[] = $val;
            }
            
            if ($status == 'NEW') {
                $this->sqlScript .= "INSERT INTO {$table} VALUES(" . implode(',', $values) . ");\n";
            } else {
                $this->sqlScript .= "REPLACE INTO {$table} VALUES(" . implode(',', $values) . ");\n";
            }
            
            $changeCount++;
        }
        
        if (!$changeCount) {
            echo "NONE\n";
        }
        
        $statement = $this->db->prepare("DROP TABLE {$tmptable}");
        $statement->execute();
        
        echo "\n";
    }
};